import { css } from '@codemirror/lang-css';
import {
  EditorView,
  WidgetType,
  ViewUpdate,
  ViewPlugin,
  DecorationSet,
  Decoration,
} from '@codemirror/view';
import { syntaxTree } from '@codemirror/language';
import { Range } from '@codemirror/rangeset';

function colorPickersDecorations(view: EditorView) {
  const widgets: Array<Range<Decoration>> = [];

  for (const range of view.visibleRanges) {
    syntaxTree(view.state).iterate({
      ...range,
      enter: (type, from, to) => {
        if (type.name === 'ColorLiteral') {
          const color = view.state.doc.sliceString(from, to);
          const d = Decoration.widget({
            widget: new ColorPickerWidget(color),
            side: 1,
          });

          widgets.push(d.range(to));
        }
      },
    });
  }

  return Decoration.set(widgets);
}

function toRRGGBBHex(color: string) {
  if (color.startsWith('#')) {
    if (color.length === 7) {
      return color;
    }

    if (color.length === 4) {
      // Convert #rgb to #rrggbb
      return `#${color[1].repeat(2)}${color[2].repeat(2)}${color[3].repeat(2)}`;
    }
  }

  return color;
}

export const wrapperClassName = 'cm-css-color-picker-wrapper';

class ColorPickerWidget extends WidgetType {
  constructor(readonly color: string) {
    super();
  }

  eq(other: ColorPickerWidget) {
    return other.color === this.color;
  }

  toDOM() {
    const picker = document.createElement('input');
    picker.type = 'color';
    // Color picker only accepts long hex
    picker.value = toRRGGBBHex(this.color);

    const wrapper = document.createElement('span');
    wrapper.appendChild(picker);
    wrapper.className = wrapperClassName;

    return wrapper;
  }

  ignoreEvent() {
    return false;
  }
}

export const colorPicker = ViewPlugin.fromClass(
  class {
    decorations: DecorationSet;

    constructor(view: EditorView) {
      this.decorations = colorPickersDecorations(view);
    }

    update(update: ViewUpdate) {
      if (update.docChanged || update.viewportChanged) {
        this.decorations = colorPickersDecorations(update.view);
      }
    }
  },
  {
    decorations: (v) => v.decorations,
    eventHandlers: {
      change: (e, view) => {
        const target = e.target as HTMLInputElement;
        if (
          target.nodeName !== 'INPUT' ||
          !target.parentElement ||
          !target.parentElement.classList.contains(wrapperClassName)
        ) {
          return false;
        }

        const pos = view.posAtDOM(target);
        const strAtPos = view.state.doc.sliceString(Math.max(0, pos - 7), pos);
        const currentColor = strAtPos.slice(strAtPos.indexOf('#'));

        if (!currentColor) {
          return false;
        }

        view.dispatch({
          changes: {
            from: pos - currentColor.length,
            to: pos,
            insert: target.value,
          },
        });

        return true;
      },
    },
  }
);
